# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Matt Fontaine

from fontTools.misc.py23 import *
from . import E_B_L_C_

class table_C_B_L_C_(E_B_L_C_.table_E_B_L_C_):

	dependencies = ['CBDT']
